#include <string.h>
#include <libfirm/firm.h>

/* struct A     { virtual int f(); };
 * struct B : A { virtual int f(); };
 * int myfunc(A* a)
 * {
 *   return a->f();
 * }
 */
int main(void)
{
	/* Bookkeeping: Initialise libFirm. */
	firm_parameter_t params;
	memset(&params, 0, sizeof(params));
	params.size = sizeof(params);
	ir_init(&params);

	/* Create an integer type. */
	ir_type* const type_int = new_type_primitive(mode_Is);

	/* Create the type struct A. */
	ident*     const id_A   = new_id_from_str("A");
	ir_type*   const type_A = new_type_class(id_A);

	/* Create the type A*. */
	ir_type* const type_pA = new_type_pointer(type_A);

	/* Add its method f(). */
	ir_type* const func_type_A = new_type_method(1 /* params */, 1 /* result */);
	set_method_param_type(func_type_A, 0, type_pA);
	set_method_res_type(func_type_A, 0, type_int);
	ident*     const id_f    = new_id_from_str("f");
	ir_entity* const ent_A_f = new_entity(type_A, id_f, func_type_A);

	/* Create the type struct B. */
	ident*     const id_B    = new_id_from_str("B");
	ir_type*   const type_B  = new_type_class(id_B);
	/* A is the supertype of B. */
	add_class_supertype(type_B, type_A);

	/* Create the type B*. */
	ir_type* const type_pB = new_type_pointer(type_B);

	/* Add its method f(). */
	ir_type* const func_type_B = new_type_method(1 /* params */, 1 /* result */);
	set_method_param_type(func_type_B, 0, type_pB);
	set_method_res_type(func_type_B, 0, type_int);
	ir_entity* const ent_B_f = new_entity(type_B, id_f, func_type_B);

	/* B::f() overwrites A::f(). */
	add_entity_overwrites(ent_B_f, ent_A_f);

	int const PARAM_a = 0;

	/* Create the type of the method. */
	ir_type* const method_type = new_type_method(1 /* param */, 1 /* result */);
	set_method_param_type(method_type, PARAM_a, type_pA);
	set_method_res_type(  method_type, 0,       type_int);

	/* Create the method entity in the global type, i.e. a normal function. */
	ident*     const id          = new_id_from_str("myfunc");
	ir_type*   const global_type = get_glob_type();
	ir_entity* const entity      = new_entity(global_type, id, method_type);
	set_entity_ld_ident(entity, id);

	/* Create the graph of the method. */
	ir_graph* const irg = new_ir_graph(entity, 0 /* local variables */);
	current_ir_graph = irg;

	/* Get the parameter. */
	ir_node* const start_block = get_irg_start_block(irg);
	ir_node* const args        = get_irg_args(irg);
	ir_node* const a           = new_r_Proj(args, mode_P_data, PARAM_a);

	/* Mature the automatically created block, it gets no further predecessors. */
	mature_immBlock(get_cur_block());

	/* Select the method f from the parameter a. */
	ir_node* const no_mem = get_irg_no_mem(irg);
	ir_node* const sel_f  = new_simpleSel(no_mem, a, ent_A_f);

	/* Call the method f(). */
	ir_node* const mem_0       = get_store();
	ir_node*       call_args[] = { a };
	ir_node* const call        = new_Call(mem_0, sel_f, 1, call_args, func_type_A);
	ir_node* const res_T       = new_Proj(call,  mode_T, pn_Call_T_result);
	ir_node* const res_0       = new_Proj(res_T, mode_Is, 0);
	ir_node* const mem_f       = new_Proj(call,  mode_M, pn_Call_M);
	set_store(mem_f);

	/* Create the return. */
	ir_node* const mem    = get_store();
	ir_node*       vals[] = { res_0 };
	ir_node* const ret    = new_Return(mem, 1, vals);

	/* Connect the return to the end block */
	ir_node* const end_block = get_irg_end_block(irg);
	add_immBlock_pred(end_block, ret);
	mature_immBlock(end_block);

	/* Bookkeeping: Set the size of the function's frame to zero. We don't need
	 * memory for local entities (memory backed variables) here. */
	ir_type* const frame_type = get_irg_frame_type(irg);
	set_type_size_bytes(     frame_type, 0);
	set_type_alignment_bytes(frame_type, 4);
	set_type_state(          frame_type, layout_fixed);

	irp_finalize_cons();

	/* Sanity check the graph. */
	irg_verify(irg, VERIFY_ENFORCE_SSA);

	/* Dump all types. */
	FILE *types = fopen("types.vcg", "w");
	dump_typegraph(types);
	fclose(types);
	/* Dump the graph. */
	FILE *code = fopen("code.vcg", "w");
	dump_ir_graph_file(code, irg);
	fclose(code);
}
