#include <string.h>
#include <libfirm/firm.h>

/* int min(int a, int b)
 * {
 *   int x;
 *   if (a < b)
 *     x = a;
 *   else
 *     x = b;
 *   return x;
 * }
 */
int main(void)
{
	/* Bookkeeping: Initialise libFirm. */
	firm_parameter_t params;
	memset(&params, 0, sizeof(params));
	params.size = sizeof(params);
	ir_init(&params);

	int const PARAM_A = 0;
	int const PARAM_B = 1;
	int const VAR_X   = 0;

	/* Create an integer type. */
	ir_type* const type_int = new_type_primitive(mode_Is);

	/* Create the type of the method. */
	ir_type* const type_method = new_type_method(2 /* params */, 1 /* result */);
	set_method_param_type(type_method, PARAM_A, type_int);
	set_method_param_type(type_method, PARAM_B, type_int);
	set_method_res_type(  type_method, 0,       type_int);

	/* Create the method entity in the global type, i.e. a normal function. */
	ident*     const id          = new_id_from_str("min");
	ir_type*   const global_type = get_glob_type();
	ir_entity* const entity      = new_entity(global_type, id, type_method);
	set_entity_ld_ident(entity, id);

	/* Create the graph of the method. */
	ir_graph* const irg = new_ir_graph(entity, 1 /* local variable */);
	current_ir_graph = irg;

	/* Get the parameters. */
	ir_node* const start_block = get_irg_start_block(irg);
	ir_node* const args        = get_irg_args(irg);
	ir_node* const a           = new_r_Proj(args, mode_Is, PARAM_A);
	ir_node* const b           = new_r_Proj(args, mode_Is, PARAM_B);

	/* Mature the automatically created block, it gets no further predecessors. */
	mature_immBlock(get_cur_block());
	/* Create the comparison. */
	ir_node* const less       = new_Cmp(a, b, ir_relation_less);
	ir_node* const cond       = new_Cond(less);
	ir_node* const true_proj  = new_Proj(cond, mode_X, pn_Cond_true);
	ir_node* const false_proj = new_Proj(cond, mode_X, pn_Cond_false);

	/* Create the then block. */
	ir_node* const then_block = new_immBlock();
	add_immBlock_pred(then_block, true_proj);
	mature_immBlock(then_block);
	set_cur_block(then_block);
	/* Set the new value for variable x in this block. */
	set_value(VAR_X, a);
	ir_node* const then_jmp   = new_Jmp();

	/* Create the else block. */
	ir_node* const else_block = new_immBlock();
	add_immBlock_pred(else_block, false_proj);
	mature_immBlock(else_block);
	set_cur_block(else_block);
	/* Set the new value for variable x in this block. */
	set_value(VAR_X, b);
	ir_node* const else_jmp   = new_Jmp();

	/* Create the block after the if with the return. */
	ir_node* const ret_block = new_immBlock();
	add_immBlock_pred(ret_block, then_jmp);
	add_immBlock_pred(ret_block, else_jmp);
	mature_immBlock(ret_block);
	set_cur_block(ret_block);
	/* Create the return. */
	ir_node* const mem    = get_store();
	ir_node* const x      = get_value(VAR_X, mode_Is);
	ir_node*       vals[] = { x };
	ir_node* const ret    = new_Return(mem, 1, vals);

	/* Connect the return to the end block */
	ir_node* const end_block = get_irg_end_block(irg);
	add_immBlock_pred(end_block, ret);
	mature_immBlock(end_block);

	/* Bookkeeping: Set the size of the function's frame to zero. We don't need
	 * memory for local entities (memory backed variables) here. */
	ir_type* const frame_type = get_irg_frame_type(irg);
	set_type_size_bytes(     frame_type, 0);
	set_type_alignment_bytes(frame_type, 4);
	set_type_state(          frame_type, layout_fixed);

	irp_finalize_cons();

	/* Sanity check the graph. */
	irg_verify(irg, VERIFY_ENFORCE_SSA);

	/* Dump the graph. */
	FILE *code = fopen("code.vcg", "w");
	dump_ir_graph_file(code, irg);
	fclose(code);
}
