#include <string.h>
#include <libfirm/firm.h>

/* struct X { int i; };
 * int reset(X* x)
 * {
 *   int i = x->i;
 *   x->i = 0;
 *   return i;
 * }
 */
int main(void)
{
	/* Bookkeeping: Initialise libFirm. */
	firm_parameter_t params;
	memset(&params, 0, sizeof(params));
	params.size = sizeof(params);
	ir_init(&params);

	int const PARAM_X = 0;

	/* Create an integer type. */
	ir_type* const type_int = new_type_primitive(mode_Is);

	/* Create the type struct X. */
	ident*     const id_X    = new_id_from_str("X");
	ir_type*   const type_X  = new_type_class(id_X);
	/* Add its entry i. */
	ident*     const id_X_i  = new_id_from_str("i");
	ir_entity* const ent_X_i = new_entity(type_X, id_X_i, type_int);
	set_entity_offset(               ent_X_i, 0);
	set_entity_offset_bits_remainder(ent_X_i, 0); // For the MiniJava types the last param is always 0.
	/* Finish the type. */
	set_type_size_bytes(     type_X, 4);
	set_type_alignment_bytes(type_X, 4);
	set_type_state(          type_X, layout_fixed);

	/* Create the type X*. */
	ir_type* const type_pX = new_type_pointer(type_X);

	/* Create the type of the method. */
	ir_type* const method_type = new_type_method(1 /* param */, 1 /* result */);
	set_method_param_type(method_type, PARAM_X, type_pX);
	set_method_res_type(  method_type, 0,       type_int);

	/* Create the method entity in the global type, i.e. a normal function. */
	ident*     const id          = new_id_from_str("reset");
	ir_type*   const global_type = get_glob_type();
	ir_entity* const entity      = new_entity(global_type, id, method_type);
	set_entity_ld_ident(entity, id);

	/* Create the graph of the method. */
	ir_graph* const irg = new_ir_graph(entity, 1 /* local variable */);
	current_ir_graph = irg;

	/* Get the parameter. */
	ir_node* const start_block = get_irg_start_block(irg);
	ir_node* const args        = get_irg_args(irg);
	ir_node* const x           = new_r_Proj(args, mode_P_data, PARAM_X);

	/* Mature the automatically created block, it gets no further predecessors. */
	mature_immBlock(get_cur_block());

	/* Select the entity i from the parameter x. */
	ir_node* const no_mem = get_irg_no_mem(irg);
	ir_node* const sel_i  = new_simpleSel(no_mem, x, ent_X_i);

	/* Load the current value x->i. */
	ir_node* const mem_0   = get_store();
	ir_node* const load_i  = new_Load(mem_0, sel_i, mode_Is, cons_none);
	ir_node* const val_i   = new_Proj(load_i, mode_Is, pn_Load_res);
	ir_node* const mem_l_i = new_Proj(load_i, mode_M,  pn_Load_M);
	set_store(mem_l_i);

	/* Set the new value x->i = 0. */
	ir_tarval* const value_zero = get_tarval_null(mode_Is);
	ir_node*   const zero       = new_Const(value_zero);
	ir_node*   const mem_1      = get_store();
	ir_node*   const store_i    = new_Store(mem_1, sel_i, zero, cons_none);
	ir_node*   const mem_s_i    = new_Proj(store_i, mode_M, pn_Store_M);
	set_store(mem_s_i);

	/* Create the return. */
	ir_node* const mem    = get_store();
	ir_node*       vals[] = { val_i };
	ir_node* const ret    = new_Return(mem, 1, vals);

	/* Connect the return to the end block */
	ir_node* const end_block = get_irg_end_block(irg);
	add_immBlock_pred(end_block, ret);
	mature_immBlock(end_block);

	/* Bookkeeping: Set the size of the function's frame to zero. We don't need
	 * memory for local entities (memory backed variables) here. */
	ir_type* const frame_type = get_irg_frame_type(irg);
	set_type_size_bytes(     frame_type, 0);
	set_type_alignment_bytes(frame_type, 4);
	set_type_state(          frame_type, layout_fixed);

	irp_finalize_cons();

	/* Sanity check the graph. */
	irg_verify(irg, VERIFY_ENFORCE_SSA);

 	/* Dump all types. */
	FILE *types = fopen("types.vcg", "w");
	dump_typegraph(types);
	fclose(types);
	/* Dump the graph. */
	FILE *code = fopen("code.vcg", "w");
	dump_ir_graph_file(code, irg);
	fclose(code);
}
