#include <string.h>
#include <libfirm/firm.h>

/* int fac(int n)
 * {
 *   int x = 1;
 *   while (n != 0)
 *   {
 *     x *= n;
 *     --n;
 *   }
 *   return x;
 * }
 */
int main(void)
{
	/* Bookkeeping: Initialise libFirm. */
	firm_parameter_t params;
	memset(&params, 0, sizeof(params));
	params.size = sizeof(params);
	ir_init(&params);

	int const PARAM_N = 0;
	int const VAR_N   = 0;
	int const VAR_X   = 1;

	/* Create an integer type. */
	ir_type* const type_int = new_type_primitive(mode_Is);

	/* Create the type of the method. */
	ir_type* const type_method = new_type_method(1 /* param */, 1 /* result */);
	set_method_param_type(type_method, PARAM_N, type_int);
	set_method_res_type(  type_method, 0,       type_int);

	/* Create the method entity in the global type, i.e. a normal function. */
	ident*     const id          = new_id_from_str("fac");
	ir_type*   const global_type = get_glob_type();
	ir_entity* const entity      = new_entity(global_type, id, type_method);
	set_entity_ld_ident(entity, id);

	/* Create the graph of the method. */
	ir_graph* const irg = new_ir_graph(entity, 2 /* local variables */);
	current_ir_graph = irg;

	/* Mature the automatically generated block. */
	mature_immBlock(get_cur_block());
	/* Get the parameter and set its variable. */
	ir_node* const start_block = get_irg_start_block(irg);
	ir_node* const args        = get_irg_args(irg);
	ir_node* const param_n     = new_r_Proj(args, mode_Is, PARAM_N);
	set_value(VAR_N, param_n);
	/* Initialise x with 1. */
	ir_tarval* const value_one = get_tarval_one(mode_Is);
	ir_node*   const one       = new_Const(value_one);
	set_value(VAR_X, one);
	ir_node* const pre_jmp = new_Jmp();

	/* Create the block with the loop header. */
	ir_node* const head_block = new_immBlock();
	add_immBlock_pred(head_block, pre_jmp);
	set_cur_block(head_block);
	/* Important: Keep one block of the loop!
	 * If the loop turns out to be an endless loop it would vanish, if it is not
	 * kept, because it is not reachable from the end anymore.
	 * Note: If there is memory used in the loop, keep one memory node (usually
	 * the Phi in the loop header), too. */
	keep_alive(head_block);
	/* Create the comparison. */
	ir_tarval* const value_zero = get_tarval_null(mode_Is);
	ir_node*   const zero       = new_Const(value_zero);
	ir_node*   const n_head     = get_value(VAR_N, mode_Is);
	ir_node*   const unequal    = new_Cmp(n_head, zero, ir_relation_less_greater);
	ir_node*   const cond       = new_Cond(unequal);
	ir_node*   const true_proj  = new_Proj(cond, mode_X, pn_Cond_true);
	ir_node*   const false_proj = new_Proj(cond, mode_X, pn_Cond_false);

	/* Create the loop body. */
	ir_node* const body_block = new_immBlock();
	add_immBlock_pred(body_block, true_proj);
	mature_immBlock(body_block);
	set_cur_block(body_block);
	/* x *= n; */
	ir_node* const x_body   = get_value(VAR_X, mode_Is);
	ir_node* const n_body   = get_value(VAR_N, mode_Is);
	ir_node* const mul      = new_Mul(x_body, n_body, mode_Is);
	set_value(VAR_X, mul);
	/* --n; */
	ir_node* const sub      = new_Sub(n_body, one, mode_Is);
	set_value(VAR_N, sub);
	/* Jump from the body back into the loop header. */
	ir_node* const body_jmp = new_Jmp();
	add_immBlock_pred(head_block, body_jmp);

	/* Now the loop header knows all its predecessors and these predecessors have
	 * all their instructions, so mature the loop header. */
	mature_immBlock(head_block);

	/* Create the block after the loop with the return. */
	ir_node* const ret_block = new_immBlock();
	add_immBlock_pred(ret_block, false_proj);
	mature_immBlock(ret_block);
	set_cur_block(ret_block);
	/* Create the return. */
	ir_node* const mem    = get_store();
	ir_node* const x_ret  = get_value(VAR_X, mode_Is);
	ir_node*       vals[] = { x_ret };
	ir_node* const ret    = new_Return(mem, 1, vals);

	/* Connect the return to the end block */
	ir_node* const end_block = get_irg_end_block(irg);
	add_immBlock_pred(end_block, ret);
	mature_immBlock(end_block);

	/* Bookkeeping: Set the size of the function's frame to zero. We don't need
	 * memory for local entities (memory backed variables) here. */
	ir_type* const frame_type = get_irg_frame_type(irg);
	set_type_size_bytes(     frame_type, 0);
	set_type_alignment_bytes(frame_type, 4);
	set_type_state(          frame_type, layout_fixed);

	irp_finalize_cons();

	/* Sanity check the graph. */
	irg_verify(irg, VERIFY_ENFORCE_SSA);

	/* Dump the graph. */
	FILE *code = fopen("code.vcg", "w");
	dump_ir_graph_file(code, irg);
	fclose(code);
}
